/*
  Args.h : For processing command line input arguments
  Copyright (C) 2013, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "Args.h"

#include "ytLib.h"
#include "util/ytParseArgs.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef USE_MPI
#include <mpi.h>
#endif

#ifdef USE_BNRC
#include "signlib.h"
#endif

void Args_help();

Args * Args_parse(int argc, char * argv[]){
  Args * args = malloc(sizeof(Args));
  int scoreMem = -1;
  ytParseArgs argset[] = {
    {"-o",                    ytParseArgs_STRING, &args->outputFile},
    {"-a",                    ytParseArgs_INT,    &args->algo},
    {"--log",                 ytParseArgs_STRING, &args->logFile},
    {"-L",                    ytParseArgs_INT,    &args->logMode},
    {"-c",                    ytParseArgs_INT,    &args->nCat},
    {"-g",                    ytParseArgs_DOUBLE, &args->g},
    {"--cth",                 ytParseArgs_INT,    &args->cth},
    {"-e",                    ytParseArgs_INT,    &args->minElms},
    {"--dry",                 ytParseArgs_INT,    &args->dry_p},
    {"--dry-procs",           ytParseArgs_INT,    &args->dry_size},
    {"--dry-threads",         ytParseArgs_INT,    &args->dry_threads},
    {"-S",                    ytParseArgs_STRING, &args->scoreOpt},
    {"--score-mem",           ytParseArgs_INT,    &scoreMem},
    {"--help",                ytParseArgs_FUNC,   Args_help},
    {ytParseArgs_DEFAULT_KEY, ytParseArgs_STRING, &args->inputFile},
    {NULL},
  };

  /*-------------------------------------------------------------------------*/
  /* Set the default arguments.                                              */
  /*-------------------------------------------------------------------------*/
  args->inputFile = NULL;
  args->outputFile = NULL;
#ifdef USE_BNRC
  args->outputFileType = "SGN3";
  args->algo = 9; /* V9 */
#else
  args->outputFileType = NULL;
  args->algo = 9;
#endif
  args->logFile = NULL;
  args->logMode = 0;
  args->maxParents = -1; /* no restriction */
  args->MPI_rank = 0;
  args->MPI_size = 1;
  args->nCat = 2;
  args->dry_p = 0;
  args->dry_size = 1;
  args->dry_threads = -1;
  args->g = 1.5;
  args->cth = 2;
  args->minElms = 100;
  args->scoreOpt = NULL;
  args->scoreMem = 1 * 1024 * 1024 * 1024; /* 1 GB by default. */

  int r = ytParseArgs_set(argset, argc, argv);
  if(r > 0){
    fprintf(stderr, "Argument Error: %s\n", argv[r]);
    exit(1);
  }

#ifdef USE_MPI
  args->MPI_comm = MPI_COMM_WORLD;
  MPI_Comm_rank(MPI_COMM_WORLD, &args->MPI_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &args->MPI_size);
#endif

  /*-------------------------------------------------------------------------*/
  /* Log Mode Setting                                                        */
  /*-------------------------------------------------------------------------*/
  switch(args->logMode){
  case 0: /* Only root */
	if(args->MPI_rank != 0){
	  ytLib_ERR = fopen("/dev/null", "w");
	} else if(args->logFile != NULL){
	  ytLib_ERR = fopen(args->logFile, "w");
	} else {
	  ytLib_ERR = stderr;
	}
	break;
  case 1: /* Every process to separate files */
	{
	  if(args->logFile == NULL) args->logFile = "log";
	  char buff[256];
	  sprintf(buff, "%s.%05d", args->logFile, args->MPI_rank);
	  FILE * fp = fopen(buff, "w");
	  ytLib_ERR = fp;
	}
	break;
  }

#ifdef USE_BNRC
  SIGN_ERR = ytLib_ERR;
#endif

  return args;
}/* Args_parse() =========================================================== */

void Args_help(){
  fprintf(ytLib_ERR,
          "ParaOS - Parallel Optimal Directed Acyclic Graph Search Software.\n"
          "USAGE: (mpirun|mpiexec) [MPI options] paraos [options] <input>\n"
          "OPTIONS:\n"
          "  -o file     Output file.\n"
          "  -c n        Number of categories in input data. (default: n=2)\n"
          "  --log file  Log file.\n"
          "  -L n        Log mode. (default: n=0)\n"
          "                0: Only root rank,  1: All ranks\n"
          );
  exit(0);
}/* Args_help() ============================================================ */

/* End of file. */
