/*
  Util.c : Utility routines
  Copyright (C) 2016, Yoshinori Tamada <tamada A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "Util.h"
#include <string.h>

/*!\brief Split jobs by the block distribution.
 *
 * This distributes 'total' jobs into 'size' pieces.
 * Simultaneously returns the first index (inclusive), the last 
 * index (exclusive), and the size of the split jobs for the given rank.
 *
 *\param total
 *\param size
 *\param rank
 *\param quot
 *\param rem
 *\param bg
 *\param ed
 *\param elm
 */
void Util_splitJobs(const uint64_t total, const int size, const int rank, 
                    uint64_t * quot, uint64_t * rem, 
                    uint64_t * bg, uint64_t * ed, uint64_t * elm){
  *quot = total / size;
  *rem = total % size;
  if(rank < (int) *rem){
	*bg = (uint64_t) ((*quot + 1) * rank);
	*ed = (uint64_t) (*bg + *quot + 1);
	*elm = *quot + 1;
  } else {
	*bg = (uint64_t) ((*quot + 1) * *rem + *quot * (rank - *rem));
	*ed = (uint64_t) (*bg + *quot);
	*elm = *quot;
  }
}
/* Util_splitJobs() ======================================================== */

void Util_splitJobs2(const uint64_t total, const int rank, 
                     const uint64_t quot, const uint64_t rem,
                     uint64_t * bg, uint64_t * ed, uint64_t * elm){
  if(rank < (int) rem){
	*bg = (uint64_t) ((quot + 1) * rank);
	*ed = (uint64_t) (*bg + quot + 1);
	*elm = quot + 1;
  } else {
	*bg = (uint64_t) ((quot + 1) * rem + quot * (rank - rem));
	*ed = (uint64_t) (*bg + quot);
	*elm = quot;
  }
}
/* Util_splitJobs() ======================================================== */


/*!\brief Returns the rank corresoponding to the given job ID.
 *\param quot
 *\param rem
 *\param id
 *\ingroup Util */
int Util_splitJobsRank(const uint64_t quot, const uint64_t rem,
                       const uint64_t id){
  if(rem == 0){
    if(quot == 0){
      return 0;
    } else {
      return id / quot;
    }
  } else {
	uint64_t a = id / (quot + 1);
	if(a < rem){
	  return a;
	} else {
	  return ((id - ((quot + 1) * rem)) / quot) + rem;
	}
  }
}/* Util_splitJobsRank() =================================================== */

void Util_fprint_ar(FILE * fp, const int * ar, const size_t len){
  size_t i;
  if(len == 0) return;
  fprintf(fp, "%d", ar[0]);
  for(i = 1; i < len; i++) fprintf(fp, ",%d", ar[i]);
}

float Util_readr_float(const uint8_t * buff){
  float f;
  //memcpy(&f, buff, sizeof(float));
  *((uint8_t *) &f)       = *(buff);
  *(((uint8_t *) &f) + 1) = *(buff + 1);
  *(((uint8_t *) &f) + 2) = *(buff + 2);
  *(((uint8_t *) &f) + 3) = *(buff + 3);
  return f;
}/* Util_readr_float() ===================================================== */

uint8_t * Util_write_int32(uint8_t * buff, int32_t value){
  memcpy(buff, &value, sizeof(int32_t));
  return buff + sizeof(int32_t);
}/* Util_write_int32() ===================================================== */

const uint8_t * Util_read_int32(const uint8_t * buff, int32_t * value){
  memcpy(value, buff, sizeof(int32_t));
  return buff + sizeof(int32_t);
}/* Util_read_int32() ====================================================== */

int32_t Util_readr_int32(const uint8_t * buff){
  uint32_t r;
  memcpy(&r, buff, sizeof(int32_t));
  return r;
}/* Util_read_int32() ====================================================== */

int8_t Util_readr_int8(const uint8_t * buff){
  int8_t r;
  memcpy(&r, buff, sizeof(int8_t));
  return r;
}/* Util_readr_int8() ======================================================= */


/* End of file. */
