/*
  io/ytFile.{h,c} : File related utility routines
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "ytFile.h"

#include <stdlib.h>
#include <string.h>

#define ytFile_BUFF_SIZE 1024

/*!\brief Reads one line from the given file stream.
 *
 * This automatically extends the size of the buffer required to
 * read the entire line.  To do so, user needs to pass the buffer and
 * the size of it by pointers.
 * The buffer can be empty (NULL).  If so, set 0 to the variable pointed
 * by \a size.
 *
 * The newline at the end of line is removed.
 *
 * This regards all CR, CR+LF, and LF as a return character.
 *
 *\param fp  input stream to be read.
 *\param buff pointer to the pointer to the buffer.  The buffer
 *  needs to be the pointer returned by malloc and so on.
 *\param size pointer to the buffer size.
 *\return 0 if succeeded, or non-zero if failed or reached to the end of
 *  the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup ytFile */
int ytFile_readLine(FILE * fp, char ** buff, size_t * size){
  char work[ytFile_BUFF_SIZE];
  size_t c = 0;
  size_t prev = 0;
  int ch;

  while((ch = getc(fp)) != EOF){
	if(ch == 10){ /* LF */
	  break;
	} else if(ch == 13) { /* CR */
	  /* Try to read one more character. */
	  ch = getc(fp);
	  if(ch != 10) ungetc(ch, fp);
	  break;
	} else {
	  if(c >= ytFile_BUFF_SIZE - 1){
		/* working buffer full.  move them to the user buffer. */
		
		if(*size < c + prev){
		  char * new_ptr;
		  new_ptr = realloc(*buff, sizeof(char) * (c + prev));
		  *buff = new_ptr;
		  *size = c + prev;
		}
		memcpy(*buff + prev, work, sizeof(char) * c);
		prev += c;
		c = 0;
	  }
	  work[c++] = (char) ch;
	}
  }/* while() */
  
  if(feof(fp) && (c + prev) == 0){
	return 1;
  }

  /* append \0 at the end of the string. */
  work[c++] = '\0';
  
  if(*size < c + prev){
	char * new_ptr;
	new_ptr = realloc(*buff, sizeof(char) * (c + prev));
	*buff = new_ptr;
	*size = c + prev;
  }
  memcpy(*buff + prev, work, sizeof(char) * c);
  return 0;
}/* ytFile_readLine() ====================================================== */

/*!\brief Prints a <TT>size_t</TT> value with delimiters for each 3-digit.
 *
 *\param fp  file stream.
 *\param value  value to be printed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup ytFile */
void ytFile_printComma_size_t(FILE * fp, size_t value){

  /* 
	 The maximum 64 bit value equals to 18,446,744,073,709,551,615.
	 Thus we need only 7 3-digit blocks to keep all digits.
   */

  int block[7];
  int i;

  for(i = 0; i < 7; i++){
	block[i] = value % 1000;
	value /= 1000;
	if(value == 0) break;
  }
  fprintf(fp, "%d", block[i]);
  for(i = i - 1; i >= 0; i--){
	fprintf(fp, ",%03d", block[i]);
  }
}
/* ytFile_printComma_size_t() ============================================== */
