/*
  math/ytPCGraph.{h,c} : Graph implementation
  Copyright (C) 2016 Yoshinori Tamada <tamada  A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __YTLIB_PC_GRAPH_H
#define __YTLIB_PC_GRAPH_H

#include <stdio.h>
#include <stdlib.h>

#ifdef DOXY
/*!\brief Value representing NULL.
 *\memberof ytPCGraph */
#define ytPCGraph_NULL
#else
#define ytPCGraph_NULL ((size_t) -1)
#endif /* DOXY */

typedef struct ytPCGraph_Edge_t {
  int pa;
  int ch;

  size_t pa_next;
  size_t ch_next;

  size_t pa_prev;
  size_t ch_prev;
} ytPCGraph_Edge;

/*!\brief Graph structure for efficient parents and children handling.
 *
 * \ref ytPCGraph is similar to \ref ytGraph except that 
 * this maintains the list of edges both for parents and
 * children.  \ref ytGraph maintains only the list of
 * parents.  Therefore, if users often need to access both
 * children and parents of certain nodes, then this is suitable
 * for that purposes.
 *
 * \see ytGraph, ytNetwork
 *
 * \author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */
typedef struct {
#ifdef DOXY
private:
#endif
  size_t nodes;
  size_t edges;
  size_t buff_size;
  ytPCGraph_Edge * ptr; /* First 'nodes' pointers are used to
						 maintain parents and children for nodes */

  /* 
	 'pa_last' & 'ch_last' maintain the last edge for each node.
	 ''pa_last[v]' ('ch_last[v]') points to the position in 'ptr'
	 of the last parent (child) node for 'v'.
   */
  size_t * pa_last; 
  size_t * ch_last;

  size_t last;
} ytPCGraph;

ytPCGraph * ytPCGraph_new(size_t nodes);
void      ytPCGraph_resize_buff(ytPCGraph * this, size_t size);
void      ytPCGraph_delete(void * this);
void      ytPCGraph_copy(const ytPCGraph * this, ytPCGraph * g);
void      ytPCGraph_clear(ytPCGraph * this);
void      ytPCGraph_status(ytPCGraph * this, FILE * fp);
void      ytPCGraph_fill(ytPCGraph * this);

size_t    ytPCGraph_num_nodes(const ytPCGraph * this);
size_t    ytPCGraph_num_edges(const ytPCGraph * this);

size_t    ytPCGraph_addEdge(ytPCGraph * this, int u, int v);
int       ytPCGraph_check_edge(ytPCGraph * this, int u, int v);
size_t    ytPCGraph_check_edge_index(ytPCGraph * this, int u, int v);
size_t    ytPCGraph_check_edge_id(ytPCGraph * this, int u, int v);
void      ytPCGraph_remove_edge_id(ytPCGraph * this, size_t id);
void      ytPCGraph_remove_edge(ytPCGraph * this, int u, int v);

size_t    ytPCGraph_first_parent_id(ytPCGraph * this, int v);
size_t    ytPCGraph_first_child_id(ytPCGraph * this, int v);
size_t    ytPCGraph_next_parent_id(ytPCGraph * this, size_t id);
size_t    ytPCGraph_next_child_id(ytPCGraph * this, size_t id);
size_t    ytPCGraph_parent_next(ytPCGraph * this, size_t id, int * pa);
size_t    ytPCGraph_child_next(ytPCGraph * this, size_t id, int * ch);

int       ytPCGraph_get_parent(ytPCGraph * this, size_t id);
int       ytPCGraph_get_child(ytPCGraph * this, size_t id);
void      ytPCGraph_get_edge(ytPCGraph * this, size_t id, int * u, int * v);
void      ytPCGraph_get_edge_index(const ytPCGraph * this, size_t index, int * u, int * v);

size_t    ytPCGraph_first_edge(ytPCGraph * this);
size_t    ytPCGraph_next_edge(ytPCGraph * this, size_t id);

int       ytPCGraph_check_cyclic(ytPCGraph * this, int u, int v);

size_t    ytPCGraph_num_parents(ytPCGraph * this, int v);
size_t    ytPCGraph_num_children(ytPCGraph * this, int v);
size_t    ytPCGraph_degree(ytPCGraph * this, int v);

void      ytPCGraph_print(ytPCGraph * this, FILE * fp);
void      ytPCGraph_print_parents(ytPCGraph * this, int j, FILE * fp);
void      ytPCGraph_print_children(ytPCGraph * this, int j, FILE * fp);

#endif /* __YTLIB_PC_GRAPH_H */

/* End of file. */
