/*
  paraos.c : Para-OS main source file
  Copyright (C) 2013-2017, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#define PARAOS_RELEASE "1.0.0"
#define PARAOS_DATE    "Mon May 22 14:55:49 2017"

#include "ytLib.h"
#include "util/ytTimer.h"
#include "io/ytFile.h"

#include "Args.h"

/* Scores  ----------------------------------------------------------------- */
#include "score.h"

/* Algorithms -------------------------------------------------------------- */
#include "os.h"

#ifdef ENABLE_V8
#include "parallel8.h"
#endif
#ifdef ENABLE_V9
#include "parallel9.h"
#endif

#ifdef _OPENMP
#include "omp.h"
#else
#include "fakeomp.h"
#endif

#include <string.h>
#include <strings.h>

typedef struct {
  int p; /* nodes */
  int n; /* samples */
  int * X; /* (nodes x samples)-matrix */
} Paraos_Data;

Paraos_Data * read_data(FILE * fp);

#ifdef USE_BNRC
#include "signlib.h"
#include "score/SGN_ScoreArgs.h"
void save_graph(SGN_Data * data, ytPCGraph * g, const char * filename, const char * type);
#endif
void test(Args * args, int p, int n, int * X, void * sb);

#ifdef HGC3
#include "util/hgc.h"
#endif

/*****************************************************************************/
/* MAIN ROUTINE                                                              */
/*****************************************************************************/
int main(int argc, char * argv[]){
  ytLib_init();
#ifdef USE_BNRC
  SGN_init();
  SIGN_ERR = ytLib_ERR;
#endif

  ytTimer * timer = ytTimer_new();
  ytTimer_start(timer);

#ifdef USE_MPI
  MPI_Init(&argc, &argv);
#endif

  Args * args = Args_parse(argc, argv);
  fprintf(ytLib_ERR, "Para-OS release " PARAOS_RELEASE " ("
          PARAOS_DATE ")\n"
          "  Copyright (C) 2013-2016 Yoshinori Tamada\n");

#ifdef PARAOS_GTIME
  {
    fprintf(ytLib_ERR, "X-GTIME:\t%d\t%d\t%d\t%.6f\n", 
            args->MPI_rank, 0, 0, ytTimer_getTime());
  }
#endif

#ifdef HGC3
  check_hgc3();
#endif

#ifdef USE_BNRC
  fprintf(ytLib_ERR, "Score: BNRC\n");
#elif defined(USE_BDE2)
  fprintf(ytLib_ERR, "Score: BDE2\n");
#elif defined(USE_BDE)
  fprintf(ytLib_ERR, "Score: BDE\n");
#endif
  fflush(ytLib_ERR);

  if(args->dry_p != 0){
    switch(args->algo){
#ifdef ENABLE_V8
    case 8:
      parallel8_mem_test(args, args->dry_p, args->dry_size);
      exit(0);
      break;
#endif
#ifdef ENABLE_V9
    case 9:
      parallel9_mem_test(args, args->dry_p, args->dry_size);
      exit(0);
      break;
#endif
    }
  }

#ifdef USE_BNRC
  SGN_Data * data;
#else
  Paraos_Data * data;
#endif

  if(args->MPI_rank == 0){
    FILE * fp = stdin;
    if(args->inputFile &&
       (fp = fopen(args->inputFile, "r")) == NULL){
      fprintf(ytLib_ERR, "ERROR: Cannot open file: %s\n", args->inputFile);
      exit(1);
    }
    if(fp == stdin){
      fprintf(ytLib_ERR, "Reading input data from the standard input...\n");
      fflush(ytLib_ERR);
    }
#ifdef USE_BNRC
    data = SGN_EDF_read_fp(fp, NULL);
#else
    data = read_data(fp);
#endif
    fclose(fp);
	fprintf(ytLib_ERR, "%d nodes, %d samples read.\n", data->p, data->n);
  } else {
    data = malloc(sizeof(Paraos_Data));
  }

  /* Broadcast the input data matrix */
#ifdef USE_MPI
#ifdef USE_BNRC
  SGN_Data_MPI_Bcast(&data, 0, args->MPI_comm);
#else
  MPI_Bcast(&data->p, 1, MPI_INT, 0, args->MPI_comm);
  MPI_Bcast(&data->n, 1, MPI_INT, 0, args->MPI_comm);
  if(args->MPI_rank != 0){
	fprintf(ytLib_ERR, "Receiving %d nodes, %d samples.\n", data->p, data->n);
    fflush(ytLib_ERR);
    data->X = malloc(sizeof(int) * data->p * data->n);
  }
  MPI_Bcast(data->X, data->p * data->n, MPI_INT, 0, args->MPI_comm);
#endif
#endif /* USE_MPI */

  if(args->maxParents == -1 || args->maxParents >= data->p){
    args->maxParents = data->p - 1;
  }

  /* Initialization */
  void ** sb;
  sb = malloc(sizeof(void *) * omp_get_max_threads());
#ifdef _OPENMP
#pragma omp parallel
#endif
  {
    int tid = omp_get_thread_num();
#ifdef USE_BNRC
    SGN_ScoreArgs * scoreArgs = SGN_ScoreArgs_news(args->scoreOpt);
    scoreArgs->max_memory = args->scoreMem;
    scoreArgs->max_parents = args->maxParents;
    scoreArgs->threads = 1;
    sb[tid] = SGN_BNRC_init(data, scoreArgs);
#elif defined(USE_BDE)
    sb[tid] = BDE_init(data->X, data->X, data->p, data->n, args->nCat,
                       args->maxParents, ytLib_ERR);
#elif defined(USE_BDE2)
    sb[tid] = BDE2_init(data->X, data->X, data->nodes, data->samples, args->nCat, 
                        args->maxParents, ytLib_ERR);
#endif
  }

  ytPCGraph * g = NULL;

  switch(args->algo){
  case 0:
    g = os(args, data->p, sb);
    break;
#ifdef ENABLE_V8
  case 8:
    g = parallel8(args, data->p, sb);
    break;
#endif
#ifdef ENABLE_V9
  case 9:
    g = parallel9(args, data->p, sb);
    break;
#endif
  case 99:
    break;
  default:
    fprintf(ytLib_ERR, "ERROR: Unknown algorithm specified. [-a %d]\n",
            args->algo);
    break;
  }

  if(g != NULL){
    ytPCGraph_print(g, ytLib_ERR);
#ifdef USE_BNRC
    save_graph(data, g, args->outputFile, args->outputFileType);
#else
    if(args->outputFile){
      FILE * fp = fopen(args->outputFile, "w");
      ytPCGraph_print(g, fp);
    }
#endif
  }

#ifdef USE_MPI
  MPI_Finalize();
#endif /* USE_MPI */

  fprintf(ytLib_ERR, "GLOBAL TIME: ");
  ytTimer_print_DDHHMMSS(timer, ytLib_ERR);
  fprintf(ytLib_ERR, "\n%f\tsec\n", ytTimer_get(timer));
  fflush(ytLib_ERR);
  return 0;
}/* main() ================================================================= */

Paraos_Data * read_data(FILE * fp){
  char * buff = NULL;
  size_t buff_size = 0;
  char * last;
  char * pos;
  char * delim;
  int flag = 1;
  int nsamples = 0;
  int line = 0;
  int * X = NULL;
  size_t mem = 0;
  while(ytFile_readLine(fp, &buff, &buff_size) == 0){
	pos = buff;
	last = index(buff, '\0');

	if(flag){
	  /* Count the number of columns in the first line. */
	  /* Note that this does not include the first column. */
	  for(int i = 0; buff[i] != '\0'; i++) if(buff[i] == '\t') nsamples++;
	  flag = 0;
	}

	/* Skip the first column */
	delim = index(pos, '\t');
	if(delim != NULL) *delim = '\0'; else delim = last;
	pos = delim + 1;

	mem += sizeof(int) * nsamples;
	X = realloc(X, mem);

	int ncol = 0;
	while(pos < last){
	  delim = index(pos, '\t');
	  if(delim != NULL) *delim = '\0'; else delim = last;
	  X[ncol + line * nsamples] = strtol(pos, NULL, 10);
	  pos = delim + 1;
	  ncol++;
	}
	if(ncol != nsamples){
	  fprintf(ytLib_ERR, "ERROR: Input file broken"
			  " at line %d (inconsistent columns).\n", line + 1);
	  exit(1);
	}
	line++;
  }
  Paraos_Data * data = malloc(sizeof(Paraos_Data));
  data->X = X;
  data->p = line;
  data->n = nsamples;
  return data;
}
/* read_data() ============================================================= */

#ifdef USE_BNRC
void save_graph(SGN_Data * data, ytPCGraph * g, 
                const char * filename, const char * type){
  IOWriter writer = IO_get_writer(type);
  if(writer == NULL){
    fprintf(ytLib_ERR, "ERROR: Unknown output file type specified: %s\n", type);
    exit(1);
  }

  SGN_Network * n = SGN_Network_new();

  for(int j = 0; j < ytPCGraph_num_nodes(g); j++){
    SGN_Node * node = SGN_Node_new(SGN_Data_get_name(data, j),
                                   SGN_Data_get_alias(data, j));
    SGN_Network_add_node(n, node);
  }

  for(int j = 0; j < ytPCGraph_num_nodes(g); j++){
    size_t id = ytPCGraph_first_parent_id(g, j);
    while(id != ytPCGraph_NULL){
      int i;
      id = ytPCGraph_parent_next(g, id, &i);

      SGN_Network_add_edge_index(n, i, j);
    }
  }
  
  FILE * outfp = stdout;
  if(filename != NULL){
    outfp = fopen(filename, "w");
    if(outfp == NULL){
      fprintf(ytLib_ERR, "WARNING: Cannot open output file: %s\n", filename);
      fprintf(ytLib_ERR, "       : Going to output in the standard output.\n");
      outfp = stdout;
    }
  }
  writer(outfp, n, NULL);
}/* save_graph() =========================================================== */
#endif

/* End of file. Copyright (C) 2016 Yoshinori Tamada */
