/*
  util/ytTimer.{h,c} : Timer functions
  Copyright (C) 2016 Yoshinori Tamada <tamada  A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "../ytLib.h"
#include "ytTimer.h"
#include "ytDoubleArray.h"
#include "ytIntArray.h"

#include <stdio.h>
#include <stdlib.h>

/*!\brief Structure for Timer functions.
 *
 * This module is to provide timer functions for measuring
 * various running times between any points in the program.
 * Users can start and stop the timer any times and can obtain
 * the current elapsed time of the timer in various formats.
 *
 * This modules uses gettimeofday() function to measure
 * times by default.
 *
 *\author Yoshinori Tamada &lt;tamada a t ytlab.jp&gt;
 *\struct ytTimer
 */

struct ytTimer_t {
  /*!\brief Start time. */
  struct timeval start;

  /*!\brief Start time in CPU clock. */
  clock_t clock_start;

  /*!\brief Accumulated time in second. */
  double sec;

  /*!\brief Whether or not currently timing. */
  int timing;

  /*!\brief Timer type */
  ytTimer_type_t type;

  /* The number of times of timing */
  int laps;

  /* lap timing measured at the latest start and stop operation. */
  double lap;

  ytDoubleArray * sectors;
  ytIntArray * lapIndex;
};

/*!\brief Constructs the \ref ytTimer instance.
 *
 * The time is measured by gettimeofday() function by default.
 * To change it, use ytTimer_type().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
ytTimer * ytTimer_new(){
  ytTimer * this = malloc(sizeof(ytTimer));
  this->sec = 0.0;
  this->timing = 0;
  this->type = ytTimer_NORMAL;
  this->laps = 0;
  this->lap = 0.0;
  this->sectors = NULL;
  this->lapIndex = NULL;
  return this;
}
/* ytTimer_new() ============================================================= */

/*!\brief Deletes the \ref ytTimer instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_delete(ytTimer * this){
  if(this->sectors != NULL){
    ytDoubleArray_delete(this->sectors);
    ytIntArray_delete(this->lapIndex);
  }
  free(this);
}
/* ytTimer_delete() ========================================================== */

void ytTimer_enableSector(ytTimer * this){
  this->sectors = ytDoubleArray_new();
  this->lapIndex = ytIntArray_new();
}/* ytTimer_enableLapping() ================================================== */

/*!\brief Sets the timer type.
 *
 * The available types are:
 * - \ref ytTimer_NORMAL - default type measured by gettimeofday() function.
 * - \ref ytTimer_CLOCK - measured by clock() function.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_set_type(ytTimer * timer, ytTimer_type_t type){
  if(type != timer->type){
	ytTimer_stop(timer);
	timer->type = type;
	ytTimer_start(timer);
  } else {
	timer->type = type;
  }
}
/* ytTimer_set_type() ==================================================== */

/*!\brief Returns the timer type.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
ytTimer_type_t ytTimer_get_type(const ytTimer * timer){
  return timer->type;
}
/* ytTimer_get_type() ==================================================== */

/*!\brief Starts to measure the time.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_start(ytTimer * timer){
  switch(timer->type){
  case ytTimer_NORMAL: gettimeofday(&timer->start, NULL); break;
  case ytTimer_CLOCK: timer->clock_start = clock(); break;
  }
  timer->timing = 1;
}
/* ytTimer_start() ======================================================= */

/*!\brief Start to measure the time with a timer type.
 *
 * This is simply an alias of calling Time_set_type() and then ytTimer_start().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_start_type(ytTimer * timer, ytTimer_type_t type){
  ytTimer_set_type(timer, type);
  ytTimer_start(timer);
}
/* ytTimer_start_type() ================================================== */

double ytTimer_timeval_span(const struct timeval * start_time,
						 const struct timeval * end_time){
  double st = 
    (double) start_time->tv_sec + ((double) start_time->tv_usec / 1000000.0);
  double ed = 
    (double) end_time->tv_sec + ((double) end_time->tv_usec / 1000000.0);

  return ed - st;
}
/* ytTimer_timeval_span() ================================================ */

/*!\brief Stops to measure the time.
 *
 * This increases the number of laps.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_stop(ytTimer * timer){
  switch(timer->type){
  case ytTimer_NORMAL:
	{
	  struct timeval end;
	  gettimeofday(&end, NULL);
	  /* If the timer is not working, just ignore */
	  if(timer->timing){
		timer->timing = 0;
		timer->lap = ytTimer_timeval_span(&timer->start, &end);

        timer->sec += timer->lap;
        if(timer->sectors){
          ytDoubleArray_add(timer->sectors, timer->lap);
        } else {
          timer->laps++;
        }
	  }
	}
	break;
  case ytTimer_CLOCK:
	{
	  clock_t clock_end = clock();
	  if(timer->timing){
		timer->timing = 0;
		timer->lap = (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
	  }
	}
	break;
  default:
	fprintf(ytLib_ERR, "ERROR: ytTimer: illegal timer type.\n");
	exit(1);
  }

}
/* ytTimer_stop() ======================================================== */

/*!\brief Stops timing and increase lapping.
 *
 */
void ytTimer_stopLap(ytTimer * this){
  ytTimer_stop(this);

  if(this->sectors){
    ytIntArray_add(this->lapIndex, (int) ytDoubleArray_size(this->sectors));
    this->laps++;
  }
}/* ytTimer_stopLap() ==================================================== */

void ytTimer_stop_no_lap(ytTimer * timer){
  double s = 0.0;
  switch(timer->type){
  case ytTimer_NORMAL:
	{
	  struct timeval end;
	  gettimeofday(&end, NULL);
	  /* If the timer is not working, just ignore */
	  if(timer->timing){
		timer->timing = 0;
		s = ytTimer_timeval_span(&timer->start, &end);
	  }
	}
	break;
  case ytTimer_CLOCK:
	{
	  clock_t clock_end = clock();
	  if(timer->timing){
		timer->timing = 0;
		s = (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
	  }
	}
	break;
  default:
	fprintf(ytLib_ERR, "ERROR: ytTimer: illegal timer type.\n");
	exit(1);
  }
  timer->lap += s;
  timer->sec += s;
}
/* ytTimer_stop() ======================================================== */

/*!\brierf Returns the lap timing.
 *
 * If the timing is not working, this returns the lap time measured between
 * the lastest start and stop operations.
 * 
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
double ytTimer_lap(ytTimer * timer){
  if(timer->timing){
	switch(timer->type){
	case ytTimer_NORMAL:
	  {
		struct timeval end;
		gettimeofday(&end, NULL);
		if(timer->timing){
		  timer->lap = ytTimer_timeval_span(&timer->start, &end);
		  gettimeofday(&timer->start, NULL);
		}
	  }
	  break;
	case ytTimer_CLOCK:
	  {
		clock_t clock_end = clock();
		if(timer->timing){
		  timer->lap = (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
		  timer->clock_start = clock();
		}
	  }
	  break;
	default:
	  fprintf(ytLib_ERR, "ERROR: ytTimer: illegal timer type.\n");
	  exit(1);
	}
	timer->sec += timer->lap;
  }
  timer->laps++;
  return timer->lap;
}
/* ytTimer_lap() ============================================================= */

void ytTimer_sector(ytTimer * this){
  if(this->timing && this->sectors){
    switch(this->type){
    case ytTimer_NORMAL:
      {
		struct timeval end;
		gettimeofday(&end, NULL);
        this->lap = ytTimer_timeval_span(&this->start, &end);
        gettimeofday(&this->start, NULL);
        ytDoubleArray_add(this->sectors, this->lap);
      }
      break;
    case ytTimer_CLOCK:
      {
        /* Not implemented */
      }
      break;
    }
    this->sec += this->lap;
  }
}/* ytTimer_sector() ======================================================= */

/*!\brief Clears the timer.
 *
 * This clears the timer.  If the timer is working, then the
 * the timer stops and is cleared.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_clear(ytTimer * timer){
  timer->timing = 0;
  timer->sec = 0.0;
}
/* ytTimer_clear() ======================================================= */

/*!\brief Clears and then starts the timer.
 *
 * This is idential to call ytTimer_clear() and then ytTimer_start().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_clearStart(ytTimer * timer){
  ytTimer_clear(timer);
  ytTimer_start(timer);
}
/* ytTimer_clearStart() ================================================= */
  
/*!\brief Returns the measured time in second.
 *
 * NOTE: This does not stop the timer.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
double ytTimer_get(const ytTimer * timer){
  switch(timer->type){
  case ytTimer_NORMAL:
	{
	  struct timeval end;
	  gettimeofday(&end, NULL);
	  if(timer->timing){
		return timer->sec + ytTimer_timeval_span(&timer->start, &end);
	  } else {
		return timer->sec;
	  }
	}
	break;
  case ytTimer_CLOCK:
	{
	  clock_t clock_end = clock();
	  if(timer->timing){
		return timer->sec + (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
	  } else {
		return timer->sec;
	  }
	}
	break;
  default:
	fprintf(ytLib_ERR, "ERROR: ytTimer: illegal timer type.\n");
	exit(1);
  }
}
/* ytTimer_get() ========================================================= */

/*!\brief Returns the average time of the laps.
 *
 * Use this after calling ytTimer_stop().
 * This returns the average time of the laps recorded until
 * the latest calling of ytTimer_stop().
 * Therefore, if the timer is currently measuring time, then
 * this does not include it for calculating the average time.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof ytTimer  */
double ytTimer_get_avg_lap(const ytTimer * timer){
  fflush(ytLib_ERR);
  return timer->sec / (double) timer->laps;
}
/* ytTimer_get_avg_lap() ================================================= */

/*!\brief Returns the absolute time (seconds from Jan. 1, 1970).
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof ytTimer  */
double ytTimer_getTime(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return (double) t.tv_sec + ((double) t.tv_usec / 1000000.0);
}/* ytTiimer_getTime() ===================================================== */

void ytTimer_print_DDHHMMSS_sec(FILE * fp, double seconds){
  int time = (int) (seconds + 0.5);
  int sec, min, hour;

  sec = time % 60;
  time /= 60;

  min = time % 60;
  time /= 60;

  hour = time % 24;
  time /= 24;

  fprintf(fp, "%d day %02d hour %02d min %02d sec",
		  time, hour, min, sec);
}
/* ytTimer_print_DDHHMMSS() ============================================== */

/*!\brief Prints the current timer in day, hour, min, sec format.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytTimer */
void ytTimer_print_DDHHMMSS(ytTimer * timer, FILE * fp){
  ytTimer_print_DDHHMMSS_sec(fp, ytTimer_get(timer));
}
/* ytTimer_print_DDHHMMSS() ============================================== */

void ytTimer_print_DDHHMMSSSSS_sec(FILE * fp, double seconds){
  int time = (int) (seconds * 1000 + 0.5);
  double sec;
  int min, hour;

  sec = (time % 60000) / 1000.0;
  time /= 60000;

  min = time % 60;
  time /= 60;

  hour = time % 24;
  time /= 24;

  fprintf(fp, "%d day %02d hour %02d min %06.3f sec",
		  time, hour, min, sec);
}
/* ytTimer_print_DDHHMMSS() ============================================== */

void ytTimer_print_DDHHMMSSSSS(const ytTimer * timer, FILE * fp){
  ytTimer_print_DDHHMMSSSSS_sec(fp, ytTimer_get(timer));
}
/* ytTimer_print_DDHHMMSS() ============================================== */

void ytTimer_print(const ytTimer * this, FILE * fp){
  if(this->sectors){
    double total = 0.0;
    int pos = 0;
    for(int i = 0; i < this->laps; i++){
      double sum = 0.0;
      for(int j = pos; j < ytIntArray_get(this->lapIndex, i); j++){
        sum += ytDoubleArray_get(this->sectors, j);
      }

      fprintf(fp, "[%d] %f (", i + 1, sum);
      int k = 0;
      for(int j = pos; j < ytIntArray_get(this->lapIndex, i); j++){
        double v = ytDoubleArray_get(this->sectors, j);
        fprintf(fp, " %d:%f", k++, v);
      }
      fprintf(fp, " )\n");
      total += sum;
      pos = ytIntArray_get(this->lapIndex, i);
    }
    fprintf(fp, "TOTAL=%f\n", total);
    fflush(fp);
  }
}/* ytTimer_print() ====================================================== */

/* End of file. */
